;; Copyright (c) 2006-2007 Carlos Ungil

;; Permission is hereby granted, free of charge, to any person obtaining
;; a copy of this software and associated documentation files (the
;; "Software"), to deal in the Software without restriction, including
;; without limitation the rights to use, copy, modify, merge, publish,
;; distribute, sublicense, and/or sell copies of the Software, and to
;; permit persons to whom the Software is furnished to do so, subject to
;; the following conditions:

;; The above copyright notice and this permission notice shall be
;; included in all copies or substantial portions of the Software.

;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :rclmath)

;; DISTRIBUTIONS

(defmacro defdist (name parameters documentation &key free (output '(r d p q)))
  `(progn 
     ,(if (member 'r output) 
	  `(defun ,(intern (concatenate 'string "R-" (symbol-name name)))
	     ,parameters
	     ,documentation
	     ,(append (list (intern (concatenate 'string "R" (symbol-name name)) "RCLMATHAPI"))
		      (loop for parameter in parameters collect `(coerce ,parameter 'double-float)))))
     ,(if (member 'd output) 
	  `(defun ,(intern (concatenate 'string "D-" (symbol-name name)))
	     ,(append (cons 'x parameters) '(&key give-log))
	     ,documentation
	     (prog1
		 ,(append (list (intern (concatenate 'string "D" (symbol-name name)) "RCLMATHAPI"))
			  '((coerce x 'double-float))
			  (loop for parameter in parameters collect `(coerce ,parameter 'double-float))
			  '((if give-log 1 0)))
	       ,(when free (list (intern (symbol-name free) "RCLMATHAPI"))))))
     ,(if (member 'p output) 
	  `(defun ,(intern (concatenate 'string "P-" (symbol-name name)))
	     ,(append (cons 'x parameters) '(&key upper-tail give-log))
	     ,documentation
	     (prog1 
		 ,(append (list (intern (concatenate 'string "P" (symbol-name name)) "RCLMATHAPI"))
			  '((coerce x 'double-float))
			  (loop for parameter in parameters collect `(coerce ,parameter 'double-float))
			  '((if upper-tail 1 0) (if give-log 1 0)))
	       ,(when free (list (intern (symbol-name free) "RCLMATHAPI"))))))
     ,(if (member 'q output) 
	  `(defun ,(intern (concatenate 'string "Q-" (symbol-name name)))
	     ,(append (cons 'x parameters) '(&key upper-tail log-p))
	     ,documentation
	     (prog1 
		 ,(append (list (intern (concatenate 'string "Q" (symbol-name name)) "RCLMATHAPI"))
			  '((coerce x 'double-float))
			  (loop for parameter in parameters collect `(coerce ,parameter 'double-float))
			  '((if upper-tail 1 0) (if log-p 1 0)))
	       ,(when free (list (intern (symbol-name free) "RCLMATHAPI"))))))))
  
;; missing p,q functions
;; handle free and output parameters

(defdist norm (mu sigma) "Normal distribution")

(defdist unif (a b) "Uniform distribution")

(defdist gamma (shape scale) "Gamma distribution")

(defdist beta (a b) "Beta distribution")

(defdist lnorm (logmean logsigma) "Lognormal distribution")
    
(defdist chisq (df) "Chi-Squared distribution")

(defdist nchisq (df lambda) "non-central Chi-Squared distribution")

(defdist f (n1 n2) "F distribution")

(defdist t (n) "Student t distribution")

(defdist binom (n p) "Binomial distribution")

(defdist nbinom (n p) "Negative Binomial distribution")

(defdist cauchy (location scale) "Chauchy distribution")

(defdist exp (scale) "Exponential distribution")

(defdist geom (p) "Geometric distribution")

(defdist hyper (NR NB n) "Hypergeometric distribution")

(defdist pois (lambda) "Poisson distribution")

(defdist weibull (shape scale) "Weibull distribution")

(defdist logis (location scale) "Logistic distribution")

(defdist wilcox (m n) "Wilcoxon Rank Sum distribution" :free wilcox-free)

(defdist signrank (n) "Wilcoxon Signed Rank distribution" :free signrank-free)

(defdist nbeta (a b lambda) "non-central Beta distribution" :output (d p q))

(defdist nf (n1 n2 ncp) "non-central F distribution" :output (d p q))

(defdist nt (df delta) "non-central Student t distribution" :output (d p q))

(defdist tukey (rr cc df) "Studentized Range distribution" :output (p q))

;; FUNCTIONS

(defun gamma (x &optional return-log)
  "Gamma function (if return-log=T the logarithm is returned)"
  (if return-log
      (rclmathapi::lgammafn (coerce x 'double-float))
  (rclmathapi::gammafn (coerce x 'double-float))))

(defun psi (x &optional (derivative 0))
  "Psi (digamma) function, or a derivative of it"
  (unless (and (integerp derivative) 
	       (not (minusp derivative)))
    (error "the order of the derivative should be a non-negative integer"))
  (case derivative
    (0 (rclmathapi::digamma (coerce x 'double-float)))
    (1 (rclmathapi::trigamma (coerce x 'double-float)))
    (2 (rclmathapi::tetragamma (coerce x 'double-float)))
    (3 (rclmathapi::pentagamma (coerce x 'double-float)))
    (t (rclmathapi::psigamma (coerce x 'double-float) 
			     (coerce derivative 'double-float)))))

(defun beta (a b &optional return-log)
  "Beta function (if return-log=T the logarithm is returned)"
  (if return-log
      (rclmathapi::lbeta (coerce a 'double-float) (coerce b 'double-float))
    (rclmathapi::beta (coerce a 'double-float) (coerce b 'double-float))))

(defun choose (n k &optional return-log)
  "number of combinations of k times chosen from n, (k and n rounded to the 
nearest integer); if return-log=T the logarithm is returned"
  (if return-log
      (rclmathapi::lchoose (coerce n 'double-float) (coerce k 'double-float))
    (rclmathapi::choose (coerce n 'double-float) (coerce k 'double-float))))

(defun bessel (type x nu &optional return-exp)
  "Bessel functions: type should be :j (:y) for the Bessel function of the first 
(second) kind and :i (:k) for the modified functions of first (second) kind; 
for the modified functions, return-exp=T results in exp(-x)I(x;nu) or exp(x)K(x; nu)"
  (unless (member type '(:j :y :i :k))
    (error "type should be one of :j (1st kind), :y (2nd kind), :i (modified 1st kind), or :k (modified 2nd kind)"))
  (when (and return-exp (member type '(:j :y)))
    (error "the return-exponential argument is only valid for modified functions"))
  (case type
    (:j (rclmathapi::bessel-j (coerce x 'double-float) (coerce nu 'double-float)))
    (:y (rclmathapi::bessel-y (coerce x 'double-float) (coerce nu 'double-float)))
    (:i (rclmathapi::bessel-i (coerce x 'double-float) (coerce nu 'double-float)
			      (if return-exp 2d0 1d0)))
    (:k (rclmathapi::bessel-k (coerce x 'double-float) (coerce nu 'double-float)
			      (if return-exp 2d0 1d0)))))
 
;; RANDOM

(defun random-normal ()
  "normally distributed pseudo-random number"
  (rclmathapi::norm-rand))

(defun random-uniform ()
  "uniformly distributed in [0,1] pseudo-random number"
  (rclmathapi::unif-rand))

(defun random-exponential ()
  "exponentially distributed pseudo-random number"
  (rclmathapi::exp-rand))

(defun rng-get-seed ()
  (cffi:with-foreign-object 
   (seed1 :int)
   (cffi:with-foreign-object 
    (seed2 :int)
    (rclmathapi::get-seed seed1 seed2)
    (list (cffi:mem-ref seed1 :int) (cffi:mem-ref seed2 :int)))))

(defun rng-set-seed (seed)
  (unless (and (integerp (first seed))
	       (integerp (second seed))
	       (null (third seed)))
    (error "the seed should be a list of two integers"))
  (rclmathapi::set-seed (first seed) (second seed))
  (rng-get-seed))

(defun rng-seed ()
  "get (or set) the seed of the random number generator (a list of two integers)"
  (rng-get-seed))

(defsetf rng-seed rng-set-seed)
